/**
 * Introduction
 *
 * @title ROV - Easy applications from C
 *
 * @copyright Jason Tribbeck / 7th ARM 1994-2001
 *
 * @authors Jason Tribbeck
 *
 * @file flex.h
 *
 * @type subsystem
 *
 * @description
 *
 * The flex subsystem is a flexible memory allocation scheme, modelled very closely
 * on Acorn's flex flexible memory allocation scheme.
 *
 * It relies on maintaining a list of memory anchors that are held in normal memory
 * that are automatically updated whenever memory is rearranged (during a resize and
 * a deallocation). These anchors form pointers to blocks of memory held in the flex
 * memory system.
 *
 * If no memory rearrangement is occurring, then pointers to memory within flex blocks
 * is permitted, but if memory rearrangement occurs, these pointers will have to be
 * recalculated based on the anchor pointers. In general, it is easier to use offsets
 * from anchors instead of pointers based on anchors. See the flex tutorial for more
 * details.
 *
 * If the non-dynamic area version is used, the normal r-versions of memory allocations
 * have to be replaced so that they use the first flex memory block (whose base address
 * never changes). This is normally by the newmalloc() functions.
 *
 * @see main.h
 */

#ifndef __rov_flex
#define __rov_flex


#ifdef __cplusplus
extern "C"
{
#endif

/**
 * Definition of a flex anchor pointer.
 */
typedef void **flex_ptr;

/**
 * Definition of a flex event handler.
 * @param type the notification type
 * @see FLEX_PRERESIZE
 * @see FLEX_POSTRESIZE
 * @see FLEX_ALLOC
 * @see FLEX_FREE
 * @see FLEX_RESIZE
 * @see FLEX_CLEAR
 * @see FLEX_SWAP
 * @see FLEX_MOVE
 * @see FLEX_USER
 */
typedef void (*flex_reg)(int type);

/**
 * Initialises the flex memory system.
 * @see flex_initdyn()
 */
void flex_init(void);

/**
 * Initialises the flex memory system using dynamic areas.
 * @see flex_init()
 * @see flex_limitdyn()
 */
void flex_initdyn(void);

/**
 * Limit the amount of dynamic area to use
 * @param size the size to limit the dynamic area to.
 * @note This must be called before calling flex_initdyn().
 * @see flex_initdyn();
 */
void flex_limitdyn(int size);

/**
 * Allocate a memory block.
 * @param anchor the pointer to the flex anchor.
 * @param size the size to allocate.
 * @return non-zero for success.
 * @note The flex anchor must exist outside of the flex memory area.
 * @see flex_allocmax()
 * @see flex_realloc()
 */
int flex_alloc(flex_ptr anchor,
	       int size);

/**
 * Allocate the most available.
 * @param anchor the pointer to the flex anchor.
 * @return non-zero for success.
 * @note The flex anchor must exist outside of the flex memory area.
 * @see flex_alloc()
 */
int flex_allocmax(flex_ptr anchor);

/**
 * Resize or allocate a flex block
 * @param anchor the flex anchor.
 * @param size the new size.
 * @return non-zero for success.
 * @note If the anchor points to NULL, then it is equivalent of
 * calling flex_alloc(). If it is not null, then it is equivalent
 * of calling flex_extend(). If the size is 0, then it is freed.
 * @see flex_alloc()
 * @see flex_extend()
 * @see flex_free()
 */
int flex_realloc(flex_ptr anchor,
		 int size);

/**
 * Frees a flex block
 * @param anchor the flex anchor.
 * @note The anchor will point to NULL after this call.
 */
void flex_free(flex_ptr anchor);

/**
 * Resize an existing flex block.
 * @param anchor the flex anchor.
 * @param size the new size.
 * @return non-zero for succes.
 * @note If the size is zero, then the anchor is freed.
 * @see flex_realloc()
 * @see flex_extendby()
 */
int flex_extend(flex_ptr anchor,
		int size);

/**
 * Resize an existing flex block relatively.
 * @param anchor the flex anchor.
 * @param extension the size to extend (or contract) by.
 * @return non-zero for success.
 * @note If the size becomes less than zero, then the anchor is freed.
 * @see flex_extend()
 */
int flex_extendby(flex_ptr anchor,
		  int extension);

/**
 * Register a flex event handler
 * @param handler the flex event handler function.
 * @note There can be only one flex event handler registered.
 */
void flex_register(flex_reg handler);

/**
 * Clears all flex allocations
 * @note all anchors will point to NULL.
 */
void flex_clear(void);

/**
 * Get the maximum allocation size.
 * @return the maximum allocation size.
 * @note This will include the WIMP's free space.
 */
int flex_max(void);

/**
 * Call the registered flex event handler.
 * @param event the event code.
 */
void flex_callregister(int event);

/**
 * Get the size of a flex block.
 * @param anchor the flex anchor.
 * @return the size of the allocated memory for that anchor.
 */
int flex_size(flex_ptr anchor);

/**
 * Swap two anchors over.
 * @param anchor1 the first anchor.
 * @param anchor2 the second anchor.
 * @note This can be seen as swapping the memory contents over, but is
 * actually swapping the pointers (and therefore very much faster).
 */
void flex_swap(flex_ptr anchor1,
	       flex_ptr anchor2);

/**
 * Moves one flex pointer to another.
 * @param anchor1 the source anchor.
 * @param anchor2 the destination anchor.
 * @note This is effectively transferring memory from one anchor to another.
 * @note If anchor2 points to existing data, then it is freed first.
 */
void flex_move(flex_ptr anchor1,
	       flex_ptr anchor2);

/**
 * Find out if flex is using dynamic areas.
 * @return non-zero if a dynamic area is being used.
 */
int flex_usingdynamic(void);

/**
 * Event code passed to flex event handler just before a resize of
 * memory is to take place.
 */
#define FLEX_PRERESIZE  (-1)

/**
 * Event code passed to flex event handler just after a resize of
 * memory has taken place.
 */
#define FLEX_POSTRESIZE (-2)

/**
 * Event code passed to flex event handler when a successful allocation
 * is about to return back to main application code.
 */
#define FLEX_ALLOC      (0)

/**
 * Event code passed to flex event handler when a successful free call
 * is about to return back to main application code.
 */
#define FLEX_FREE       (1)

/**
 * Event code passed to flex event handler when a successful resize
 * call is about to return back to main application code.
 */
#define FLEX_RESIZE     (2)

/**
 * Event code passed to flex event handler when a successful clear
 * call is about to return back to main application code. 
 */
#define FLEX_CLEAR      (3)

/**
 * Event code passed to flex event handler when a successful swap
 * call is about to return back to main application code.
 */
#define FLEX_SWAP       (4)

/**
 * Event code passed to flex event handler when a successful move
 * call is about to return back to main application code.
 */
#define FLEX_MOVE       (5)

/**
 * Event code base number for application events to pass into
 * flex event handler.
 */
#define FLEX_USER       (100)


#ifdef __cplusplus
}
#endif

#endif
